/*
 * Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that: (1) source code distributions
 * retain the above copyright notice and this paragraph in its entirety, (2)
 * distributions including binary code include the above copyright notice and
 * this paragraph in its entirety in the documentation or other materials
 * provided with the distribution, and (3) all advertising materials mentioning
 * features or use of this software display the following acknowledgement:
 * ``This product includes software developed by the University of California,
 * Lawrence Berkeley Laboratory and its contributors.'' Neither the name of
 * the University nor the names of its contributors may be used to endorse
 * or promote products derived from this software without specific prior
 * written permission.
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND WITHOUT ANY EXPRESS OR IMPLIED
 * WARRANTIES, INCLUDING, WITHOUT LIMITATION, THE IMPLIED WARRANTIES OF
 * MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE.
 */
#ifndef lint
static char copyright[] =
    "Copyright (c) 1988, 1989, 1990, 1991, 1992, 1993, 1994\n Lawrence Berkeley Laboratory. All rights reserved.\n";
static char rcsid[] =
    "@(#) $Header: leinfo.c,v 1.19 94/08/24 18:39:07 leres Exp $ (LBL)";
#endif

#include <sys/param.h>
#include <sys/buf.h>
#include <sys/ioctl.h>
#include <sys/map.h>
#include <sys/mbuf.h>
#include <sys/socket.h>
#include <sys/time.h>
#include <sys/kernel.h>

#include <net/if.h>
#include <netinet/in.h>
#include <netinet/if_ether.h>
#ifdef SUNOS4
#include <netinet/in_var.h>
#endif

#include <sundev/mbvar.h>

#include <sunif/if_lereg.h>
#include <sunif/if_levar.h>

#include <arpa/inet.h>

#include <ctype.h>
#include <errno.h>
#include <fcntl.h>
#include <kvm.h>
#include <nlist.h>
#include <stdio.h>
#include <strings.h>

#define satosin(dst) ((struct sockaddr_in *)(dst))

static struct nlist nl[] = {
#define NL_IFNET 0
	{ "_ifnet" },
	{ "" }
};

struct var {
	char *name;
	void *addr;
	int flags;
};

#define VAR_SHORT	0x1
#define VAR_IN		0x2
#define VAR_OUT		0x4

static struct le_softc softc;

static struct var softc_var[] = {
	{ "es_nrmds",		&softc.es_nrmds,	0 },
	{ "es_ntmds",		&softc.es_ntmds,	0 },
	{ "es_nrbufs",		&softc.es_nrbufs,	0 },

	{ "es_extrabyte",	&softc.es_extrabyte,	0 },
	{ "es_fram",		&softc.es_fram,		0 },
	{ "es_crc",		&softc.es_crc,		0 },
	{ "es_oflo",		&softc.es_oflo,		0 },
	{ "es_uflo",		&softc.es_uflo,		0 },
	{ "es_retries",		&softc.es_retries,	0 },
	{ "es_missed",		&softc.es_missed,	0 },
#ifdef SUNOS4
	{ "es_memerr",		&softc.es_missed,	0 },
	{ "es_init",		&softc.es_init,		0 },
#endif
	{ "es_noheartbeat",	&softc.es_noheartbeat,	0 },
	{ "es_tBUFF",		&softc.es_tBUFF,	0 },
	{ "es_tlcol",		&softc.es_tlcol,	0 },
	{ "es_trtry",		&softc.es_trtry,	0 },
	{ "es_tnocar",		&softc.es_tnocar,	0 },
#ifdef SUNOS4
	{ "es_dogreset",	&softc.es_dogreset,	0 },
#endif

	/* Performance statistics counters */
#ifdef SUNOS3
	{ "es_started",		&softc.es_started,	0 },
	{ "es_started2",	&softc.es_started2,	0 },
#endif
	{ "es_no_tmds",		&softc.es_no_tmds,	0 },
#ifdef SUNOS3
	{ "es_potential_rloans",&softc.es_potential_rloans, 0 },
	{ "es_actual_rloans",		&softc.es_actual_rloans,	0 },
#endif
	{ "es_tsync",		&softc.es_tsync,	0 },
#ifdef SUNOS4
	{ "es_defer",		&softc.es_defer,	0 },
	{ "es_outcpy",		&softc.es_outcpy,	0 },
	{ "es_incpy",		&softc.es_incpy,	0 },
#endif
	{ NULL }
};

int *pctin = &softc.es_if.if_ipackets;
int *pctout = &softc.es_if.if_opackets;

static struct var ifnet_var[] = {
	{ "if_ipackets",	&softc.es_if.if_ipackets,	0 },
	{ "if_ierrors",		&softc.es_if.if_ierrors,	VAR_IN },
	{ "if_opackets",	&softc.es_if.if_opackets,	0 },
	{ "if_oerrors",		&softc.es_if.if_oerrors,	VAR_OUT },
	{ "if_collisions",	&softc.es_if.if_collisions,	VAR_OUT },
	{ NULL }
};

static struct var ifqueue_var[] = {
	{ "ifq_len",		&softc.es_if.if_snd.ifq_len,	0 },
	{ "ifq_maxlen",		&softc.es_if.if_snd.ifq_maxlen,	0 },
	{ "ifq_drops",		&softc.es_if.if_snd.ifq_drops,	0 },
	{ NULL }
};


static char *prog;
static kvm_t *kd;
static int vflag;

void kread();
char *sprintb();
char *pct();

main(argc, argv)
	int argc;
	char **argv;
{
	register u_int v;
	register char *cp;
	register struct le_softc *sc;
	register struct var *vp;
	char *myname = "le";
	register struct ifnet *ifp;
#ifdef SUNOS4
	struct in_ifaddr ifaddr;
#endif
	u_long addr;
	char name[16];
	char *vmunix = NULL;
	char *kmem = NULL;
	int unit = 0;
	char *sfmt = "%20s: %s\n";
	char *dfmt = "%20s: %d\n";
	char *pfmt = "%20s: %-11d (%s)\n";

	if (cp = rindex(argv[0], '/'))
		prog = cp + 1;
	else
		prog = argv[0];

	if (argc > 1 && strcmp(argv[1], "-v") == 0) {
		vflag = 1;
		--argc; ++argv;
	}
	if (argc > 1 && isdigit(argv[1][0])) {
		unit = atoi(*++argv);
		--argc;
	}
	if (argc > 1)
		vmunix = argv[1];
	if (argc > 2)
		kmem = argv[2];

	if ((kd = kvm_open(vmunix, kmem, (void *)0, O_RDONLY, prog)) == 0) {
		exit(1);
		/* NOTREACHED */
	}
	if (kvm_nlist(kd, nl) < 0) {
		fprintf(stderr, "%s: kvm_nlist\n", prog);
		exit(1);
		/* NOTREACHED */
	}
	if (nl[NL_IFNET].n_type == 0) {
		fprintf(stderr, "%s: no namelist\n", prog);
		exit(1);
		/* NOTREACHED */
	}


	/* Loop through ifnet's looking for ours */
	kread(nl[NL_IFNET].n_value, (char *)&addr, sizeof(addr), "ifnet addr");
	sc = &softc;
	ifp = &sc->es_ac.ac_if;
	for (; addr != 0; addr = (u_long)ifp->if_next) {
		/* Read ifnet */
		kread(addr, (char *)ifp, sizeof(*ifp), "ifnet");
		if (ifp->if_unit != unit)
			continue;

		/* Read interface name */
		kread((u_long)ifp->if_name, name, sizeof(name), "name");
		name[sizeof(name) - 1] = '\0';
		if (strcmp(myname, name) == 0)
			break;
	}
	if (addr == 0) {
		fprintf(stderr, "%s: Can't find interface %s%d\n",
		    prog, myname, unit);
		exit(1);
		/* NOTREACHED */
	}
	/* Read softc */
	kread(addr, (char *)sc, sizeof(*sc), "softc");

	printf(dfmt, "if_unit", ifp->if_unit);
	printf(dfmt, "if_mtu", ifp->if_mtu);

#ifdef SUNOS4
/* SunOS 4 code */

	/* Get ifaddr struct */
	if (addr = (u_long)ifp->if_addrlist)
		kread(addr, (char *)&ifaddr, sizeof(ifaddr), "ifaddr");
#define NONE(s) (addr ? s : "none")

	printf(sfmt, "ia_addr",
	    NONE(inet_ntoa(satosin(&ifaddr.ia_addr)->sin_addr)));
	printf(sfmt, "ia_broadaddr",
	    NONE(inet_ntoa(satosin(&ifaddr.ia_broadaddr)->sin_addr)));

	printf(sfmt, "ia_net",
	    NONE(inet_ntoa(*(struct in_addr *)&ifaddr.ia_net)));
	printf(sfmt, "ia_netmask",
	    NONE(inet_ntoa(*(struct in_addr *)&ifaddr.ia_netmask)));
	printf(sfmt, "ia_subnet",
	    NONE(inet_ntoa(*(struct in_addr *)&ifaddr.ia_subnet)));
	printf(sfmt, "ia_subnetmask",
	    NONE(inet_ntoa(*(struct in_addr *)&ifaddr.ia_subnetmask)));
	printf(sfmt, "ia_netbroadcast",
	    NONE(inet_ntoa(ifaddr.ia_netbroadcast)));

	if (addr && ifaddr.ia_ifa.ifa_next) {
		char buf[32];

		sprintf(buf, "0x%x", ifaddr.ia_ifa.ifa_next);
		printf(sfmt, "ifa_next", buf);
	}
#define	IFFBITS \
"\020\1UP\2BROADCAST\3DEBUG\4LOOPBACK\5POINTOPOINT\6NOTRAILERS\7RUNNING\10NOARP\
\11PROMISC\12ALLMULTI\14MULTICAST\15LINK0\016LINK1\017LINK2\020PRIVATE"
#define LEBITS "\020\1PIO\2IOADDR\3TBUSY\4NEEDTMD\5DOGNEEDTMD"
#else
/* SunOS 3 code */

	printf(sfmt, "if_net", inet_ntoa(*(struct in_addr *)&ifp->if_net));

	printf(sfmt, "if_addr", inet_ntoa(satosin(&ifp->if_addr)->sin_addr));
	printf(sfmt, "if_host[0]",
	    inet_ntoa(*(struct in_addr *)&ifp->if_host[0])); 
	printf(sfmt, "if_broadaddr",
	    inet_ntoa(satosin(&ifp->if_broadaddr)->sin_addr));
	printf(sfmt, "if_netmask",
	    inet_ntoa(*(struct in_addr *)&ifp->if_netmask));
	printf(sfmt, "if_subnet",
	    inet_ntoa(*(struct in_addr *)&ifp->if_subnet));
	printf(sfmt, "if_subnetmask",
	    inet_ntoa(*(struct in_addr *)&ifp->if_subnetmask));

#define	IFFBITS \
"\020\1UP\2BROADCAST\3DEBUG\4ROUTE\5POINTOPOINT\6NOTRAILERS\7RUNNING\10NOARP\
\11PROMISC\12LOOPBACK\16LINK2\17LINK1\20LINK0"
#define LEBITS "\020\1TBUSY\2TOPENDING\3NEEDTMD"
#endif
	printf(sfmt, "if_flags", sprintb(ifp->if_flags, IFFBITS));
	if (sc->es_flags != 0)
		printf(sfmt, "es_flags", sprintb(sc->es_flags, LEBITS));

	/* Ifnet counters */ 
	for (vp = ifnet_var; vp->name != NULL; ++vp) {
		if (vp->flags & VAR_SHORT)
			v = *(u_short *)vp->addr;
		else
			v = *(u_int *)vp->addr;
		if (vp->flags & VAR_IN)
			cp = pct(v, *pctin);
		else if (vp->flags & VAR_OUT)
			cp = pct(v, *pctout);
		else
			cp = NULL;
		if (cp == NULL)
			printf(dfmt, vp->name, v);
		else
			printf(pfmt, vp->name, v, cp);
	}

	printf(dfmt, "if_timer", ifp->if_timer);

	for (vp = ifqueue_var; vp->name != NULL; ++vp)
		printf(dfmt, vp->name, *(int *)vp->addr);

	/* Softc error and performance counters */ 
	for (vp = softc_var; vp->name != NULL; ++vp) {
		if (vp->flags & VAR_SHORT)
			v = *(u_short *)vp->addr;
		else
			v = *(u_int *)vp->addr;
		if (v != 0 || vflag)
			printf(dfmt, vp->name, v);
	}

	if (vflag) {
		printf("initblk @ 0x%08x\n", sc->es_ib);
		printf("rd ring @ 0x%08x to 0x%08x, cur 0x%08x\n",
			sc->es_rdrp, sc->es_rdrend, sc->es_his_rmd);
		printf("td ring @ 0x%08x to 0x%08x, cur 0x%08x, nxt 0x%08x\n",
			sc->es_tdrp, sc->es_tdrend, sc->es_cur_tmd,
			sc->es_nxt_tmd);
	}
}

/*
 * Print a value a la the %b format of the kernel's printf
 */
char *
sprintb(v, bits)
	register char *bits;
	register unsigned short v;
{
	register char *cp;
	register int i, any = 0;
	register char c;
	static char buf[132];

	if (bits && *bits == 8)
		(void)sprintf(buf, "0%o", v);
	else
		(void)sprintf(buf, "0x%x", v);
	cp = buf + strlen(buf);
	bits++;
	if (bits) {
		*cp++ = ' ';
		*cp++ = '(';
		while (i = *bits++) {
			if (v & (1 << (i-1))) {
				if (any)
					*cp++ = ',';
				any = 1;
				for (; (c = *bits) > 32; bits++)
					*cp++ = c;
			} else
				for (; *bits > 32; bits++)
					continue;
		}
		*cp++ = ')';
	}
	*cp = '\0';
	return (buf);
}

char *
pct(v, t)
	register int v, t;
{
	static char buf[32];

	if (t == 0)
		return "NaN";
	(void)sprintf(buf, "%.1f%%", 100.0 * ((double)v) / ((double)t));
	return (buf);
}

void
kread(addr, buf, size, errstr)
	u_long addr;
	char *buf;
	int size;
	char *errstr;
{

	if (kvm_read(kd, addr, buf, size) != size) {
		fprintf(stderr, "%s: kvm_read: ", prog);
		perror(errstr);
		exit(1);
		/* NOTREACHED */
	}
}
